import java.io.File;
import java.io.FileInputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

public class NodeParser extends DefaultHandler implements Runnable {
	static long nodeCount = 0;
	static PreparedStatement pNode;
	static PreparedStatement pTag;
	static public Streamer s;
	boolean inNode = false;
	DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss'Z'");

	private long tagLength = 0;
	private long nodeLength = 0;
	public static boolean commiting = false;
	
	public NodeParser(File f) {
		try {
			Connection cn = ConnectionFactory.getNewConnection();
			pTag = cn.prepareStatement("INSERT IGNORE INTO tags (nodeid,name,value,sort) VALUES(?,?,?,?)");
			pNode = cn.prepareStatement("INSERT IGNORE INTO nodes (nodeid,lat,lon,version,timestamp) VALUES(?,?,?,?,?)");
			s = new Streamer(new FileInputStream(f));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public void run() {
		try {
			SAXParser saxParser = SAXParserFactory.newInstance().newSAXParser();
			saxParser.parse(s, this);
			
			commit();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	@Override
	public void startElement(String namespaceURI, String localName, // local
			// name
			String qName, // qualified name
			Attributes attrs) throws SAXException {
		String eName = ("".equals(localName)) ? qName : localName;
		if( "node".equals(eName) ) {
			handleNodeStart(attrs);
		} else if( "tag".equals(eName) ) {
			handleTagStart(attrs);
		}
	}
	
	private void handleNodeStart(Attributes attrs) throws SAXException {
		Node.clear();
		if (attrs != null) {
			String aName;
			String value;
			for (int i = 0; i < attrs.getLength(); i++) {
				aName = attrs.getLocalName(i); // Attr name
				if ("".equals(aName) )
					aName = attrs.getQName(i);
				
				value = attrs.getValue(i);
				if( "id".equals(aName) ) {
					Node.id = Long.parseLong(value);
				} else if( "lat".equals(aName) ) {
					Node.lat = Double.parseDouble(value);
				} else if( "lon".equals(aName) ) {
					Node.lon = Double.parseDouble(value);
				} else if( "version".equals(aName) ) {
					Node.version = Integer.parseInt(value);
				} else if( "timestamp".equals(aName) ) {
					LocalDateTime ldt = LocalDateTime.parse(value, dtf);
					Node.timestamp = Timestamp.valueOf(ldt);
				}
			}
		}
		
		inNode = true;
		nodeCount++;
	}

	private void handleTagStart(Attributes attrs) {
		if (attrs != null) {
			String k = null;
			String v = null;
			String aName;
			String value;
			for (int i = 0; i < attrs.getLength(); i++) {
				aName = attrs.getLocalName(i); // Attr name
				if ("".equals(aName) )
					aName = attrs.getQName(i);
				
				value = attrs.getValue(i);
				if( "k".equals(aName) ) {
					k = value;
				} else if( "v".equals(aName) ) {
					v = value;
				}
			}
			
			if( k!=null && v != null) {
				if( inNode ) {
					Node.tags.add(new Tag(k,v));
				}
			}
		}
	}

	@Override
	public void endElement(String namespaceURI, String localName, // local name
			String qName) // qualified name
			throws SAXException {
		String eName = ("".equals(localName)) ? qName : localName;
		if( "node".equals(eName) ) {
			inNode = false;
			
			addNode();
		}
	}
	
	private void addNode() {
		try {
			int tagIdx = 1;
			for(Tag t : Node.tags) {
				pTag.setLong(1, Node.id);
				pTag.setString(2,  t.name);
				pTag.setString(3,  t.value);
				pTag.setInt(4,  tagIdx);
				tagLength++;

				pTag.addBatch();
				tagIdx++;
			}
			pNode.setLong(1, Node.id);
			pNode.setDouble(2,  Node.lat);
			pNode.setDouble(3,  Node.lon);
			pNode.setInt(4,  Node.version);
			pNode.setTimestamp(5,  Node.timestamp);
			nodeLength++;
			
			pNode.addBatch();
			
			if( tagLength > 100000 || nodeLength > 100000) {
				commit();
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
		Node.clear();
	}
	
	protected void commit() throws SQLException {
		commiting = true;
		try {
			pTag.executeBatch();
			pTag.clearBatch();
		} catch(Exception e) {
			throw new SQLException("node-tag-exception", e);
		}

		try {
			pNode.executeBatch();
			pNode.clearBatch();
		} catch(Exception e) {
			throw new SQLException("node-exception", e);
		}
		commiting = false;
		
		tagLength = 0;
		nodeLength = 0;
	}
}
